;;; -*- lexical-binding: t -*-

(require 'eieio)

(require 'bad-box)
(require 'bad-box-draw)
(require 'bad-box-rotate)
(require 'bad-caption)
(require 'bad-elem)
(require 'bad-game)
(require 'bad-mode)
(require 'bad-write)

(defclass studio (game)
  ((name      :initform "studio")
   (canvas    :initarg :canvas    :type (or null box)  :initform nil)
   (elems     :initarg :elems     :type list           :initform nil)
   (sel       :initarg :sel       :type (or null elem) :initform nil)
   (ie        :initarg :ie        :type integer        :initform 0)
   (pos-funs  :initarg :pos-funs  :type list           :initform (list #'origo     #'center))
   (size-funs :initarg :size-funs :type list           :initform (list #'size-half #'size-square #'size-max #'size-min))))

(cl-defmethod init ((s studio))
  (with-slots (screen canvas elems sel help) s
    (setf help (caption
                :name "help"
                :text (string-join
             #3=(list #2=(format "%s help     " #1="[h]")
                                "[a] add box  "
                                "[t] triangle "
                                "[n] .. next  "
                                "[p] .. prev  "
                                "[s] box side "
                                "[c] .. corner"
                                "[v] visible  "
                                "[r] rotate   "
                                "[R] .. back  "
                                "[T] transpose"
                                "[f] flip x   "
                                "[F] .. y     "
                                "[d] size     "
                                "[D] .. full  "
                                "[;] position "
                                "[q] quit     "
                                "             "
                                "     [i]     "
                                "[j]  mov  [l]"
                                "     [k]     "
                                "             "
                                "[u]  [I]  [o]"
                                "[J]  adj  [K]"
                                "     [L]     "))
                :w #4=(length #2#) :h (length #3#)
                :text-alt (string-pad #1# #4# bad-nonsolid t)))
    (draw help)

    (with-slots (w) help
      (let ((wc 78)
            (hl 27))
        (pos help #1=(- wc w))
        (setq canvas (box
                      :name "canvas"
                      :movable nil
                      :c1 #2=?~
                      :c2 #2#
                      :c3 #2#
                      :c4 #2#
                      :s1 #3='(#2#)
                      :s2 #4='(?-)
                      :s3 #3#
                      :s4 #4#
                      :w (1- #1#)
                      :h hl))
        (draw canvas)))

    (with-slots (sub) screen
      (setf sub (list canvas help)))
    (draw screen)

    (setf elems (list canvas))
    (setf sel canvas)
    (setup-bad-mode s)))

(cl-defmethod setup-bad-mode ((s studio))
  (init-keys s)
  (pop-to-buffer (bad-mode-buffer))
  (delete-other-windows))

(cl-defmethod draw ((s studio))
  (with-slots (screen) s
    (draw screen)
    (draw-to-buf screen)))

(cl-defmethod run ((s studio))
  (init s)
  (draw s))

(cl-defmethod init-keys ((s studio))
  (setq bad-mode-map (make-keymap))
  (let ((map bad-mode-map))
    (set-char-table-range (nth 1 map) t #'ignore)

    ;; cycle
    (keymap-set map "p" (lambda () (interactive) (prev s)))
    (keymap-set map "n" (lambda () (interactive) (next s)))

    ;; rotate
    (keymap-set map "r" (lambda () (interactive) (operate s #'rotate)))
    (keymap-set map "R" (lambda () (interactive) (operate s #'rotate-back)))
    (keymap-set map "f" (lambda () (interactive) (operate s #'flip-x)))
    (keymap-set map "F" (lambda () (interactive) (operate s #'flip-y)))
    (keymap-set map "T" (lambda () (interactive) (operate s #'transpose)))

    ;; add
    (keymap-set map "a" (lambda () (interactive) (add-box      s)))
    (keymap-set map "t" (lambda () (interactive) (add-triangle s)))

    ;; sides
    (keymap-set map "c" (lambda () (interactive) (set-corner s)))
    (keymap-set map "s" (lambda () (interactive) (set-side   s)))

    ;; size
    (keymap-set map "d" (lambda () (interactive) (cycle-size s)))
    (keymap-set map "D" (lambda () (interactive) (operate s #'size-max)))

    ;; row
    (keymap-set map "I" (lambda () (interactive) (operate s #'add-row -1)))
    (keymap-set map "K" (lambda () (interactive) (operate s #'add-row)))

    ;; col
    (keymap-set map "u" (lambda () (interactive) (operate s #'add-col -1)))
    (keymap-set map "o" (lambda () (interactive) (operate s #'add-col)))
    (keymap-set map "J" (lambda () (interactive) (operate s #'add-col -1)))
    (keymap-set map "L" (lambda () (interactive) (operate s #'add-col)))

    ;; pos
    (keymap-set map ":" (lambda () (interactive) (cycle-pos s)))
    (with-slots (canvas) s
      (keymap-set map ";" (lambda () (interactive) (operate s #'pos-inside canvas))))

    ;; move
    (keymap-set map "i" (lambda () (interactive) (operate s #'up)))
    (keymap-set map "k" (lambda () (interactive) (operate s #'down)))
    (keymap-set map "j" (lambda () (interactive) (operate s #'left)))
    (keymap-set map "l" (lambda () (interactive) (operate s #'right)))

    ;; unsorted
    (keymap-set map "h" (lambda () (interactive) (help-toggle s)))
    (keymap-set map "v" (lambda () (interactive) (operate s #'visible)))
    (keymap-set map "q" #'kill-emacs)))

;; -----------------------------------------------------------------------

(cl-defmethod next ((s studio) &optional n)
  (or n (setq n 1))
  (with-slots (elems sel ie) s
      (when elems
        (setf sel (nth (mod (cl-incf ie n) (length elems)) elems))
        (when (and sel (slot-boundp sel 'name))
          (with-slots (name) sel
            (message name))))))

(cl-defmethod prev ((s studio) &optional n)
  (or n (setq n 1))
  (next s (- n)))

;; -----------------------------------------------------------------------

(cl-defmethod operate ((s studio) fun &rest args)
  (with-slots (sel) s
    (when (and sel (functionp fun))
      (apply fun `(,sel ,@args))
      (draw s))))

;; -----------------------------------------------------------------------

(cl-defmethod add ((s studio) (e elem))
  (with-slots (sel elems) s
    (when sel
      (let ((inner (add sel e)))
        (push inner elems)
        (setq sel inner)
        (draw s)))))

(cl-defmethod add-box ((s studio))
  (add s (box)))

(cl-defmethod add-triangle ((s studio))
  (add s (triangle)))

;; -----------------------------------------------------------------------

(cl-defmethod set-side ((s studio))
  (let ((side (read-number "side? (1-4) "))
        (str  (read-string (format "pattern? [%s] " #1="-*") nil nil #1#)))
    (with-slots (sel) s
      (when (and sel (box-p sel))
        (side sel side str)
        (draw s)))))

(cl-defmethod set-border ((s studio))
  (with-slots (sel) s
    (when (and sel (box-p sel))
      (border sel)
      (draw s))))

(cl-defmethod set-corner ((s studio))
  (let ((str (read-string (format "pattern? (1-4 chars) [%s] " #1="1234") nil nil #1#)))
    (with-slots (sel) s
      (when (and sel (box-p sel))
        (let* ((cs  (string-to-list str))
               (len (length cs)))
          (cl-case len
            (1 (set-corner (make-string 4 (car cs))))
            (2 (let ((cs-fst (car  cs))
                     (cs-snd (cadr cs)))
                 (set-corner
                  (format "%c%c%c%c" cs-fst cs-snd cs-fst cs-snd))))
            (3 (set-corner (format "%c%c%c%c"
                                   (nth 0 cs)
                                   (nth 1 cs)
                                   (nth 2 cs)
                                   (nth 2 cs))))
            (4 (dotimes (i 4)
                 (corner sel (1+ i) (nth i cs))))
            (_ (prog1 nil
                 (message "input 1-4 chars"))))))
      (draw s))))

;; -----------------------------------------------------------------------

(cl-defmethod cycle-size ((s studio))
  (with-slots (size-funs sel) s
    (funcall (car size-funs) sel)
    (setf size-funs (bad-list-shift size-funs)))
  (draw s))

(cl-defmethod cycle-pos ((s studio))
  (with-slots (sel pos-funs) s
    (when (and sel pos-funs)
      (with-slots (x y) sel
        (let ((old-x x)
              (old-y y))
          (funcall (car pos-funs) sel)
          (setf pos-funs (bad-list-shift pos-funs))
          (when (and (= x old-x)
                     (= y old-y))
            (funcall (car pos-funs) sel)
            (setq pos-funs (bad-list-shift pos-funs))))
        (draw s)))))

;; -----------------------------------------------------------------------

(cl-defmethod draw-to-buf ((s studio) &optional e)
  (with-slots (sel) s
    (or e (setq e sel))
    (when (cl-typep e 'elem)
      (draw-to-buf e)
      (draw s))))

(provide 'bad-studio)
